@@ -152,8 +152,10 @@ int arg_to_num(arg_struct *arg, int4 *num) {
 
 int recall_result_silently(vartype *v) {
     if (flags.f.stack_lift_disable) {
-        // sp guaranteed to be >= 0 in this case
-        free_vartype(stack[sp]);
+        if (sp == -1)
+            sp = 0;
+        else
+            free_vartype(stack[sp]);
     } else if (flags.f.big_stack) {
         if (!ensure_stack_capacity(1)) {
             free_vartype(v);
@@ -179,13 +181,14 @@ int recall_result(vartype *v) {
 
 int recall_two_results(vartype *x, vartype *y) {
     if (flags.f.big_stack) {
-        int off = flags.f.stack_lift_disable ? 1 : 2;
+        bool sld = flags.f.stack_lift_disable && sp != -1;
+        int off = sld ? 1 : 2;
         if (!ensure_stack_capacity(off)) {
             free_vartype(x);
             free_vartype(y);
             return ERR_INSUFFICIENT_MEMORY;
         }
-        if (flags.f.stack_lift_disable)
+        if (sld)
             free_vartype(stack[sp]);
         sp += off;
     } else {
@@ -745,6 +748,8 @@ int virtual_flag_handler(int flagop, int flagnum) {
 }
 
 int get_base() {
+    if (flags.f.prgm_mode)
+        return 10;
     int base = 0;
     if (flags.f.base_bit0) base += 1;
     if (flags.f.base_bit1) base += 2;
@@ -756,7 +761,7 @@ int get_base() {
         return 10;
 }
 
-void set_base(int base) {
+void set_base(int base, bool a_thru_f) {
     int oldbase = 0;
     if (flags.f.base_bit0) oldbase += 1;
     if (flags.f.base_bit1) oldbase += 2;
@@ -771,7 +776,7 @@ void set_base(int base) {
     flags.f.base_bit1 = (base & 2) != 0;
     flags.f.base_bit2 = (base & 4) != 0;
     flags.f.base_bit3 = (base & 8) != 0;
-    if (mode_appmenu == MENU_BASE_A_THRU_F)
+    if (!a_thru_f && mode_appmenu == MENU_BASE_A_THRU_F)
         set_menu(MENULEVEL_APP, MENU_BASE);
 
     if (base != oldbase)
@@ -796,6 +801,8 @@ int base_range_check(int8 *n, bool force_wrap) {
                 *n &= (1ULL << wsize) - 1;
         }
     } else if (flags.f.base_signed) {
+        if (wsize == 64)
+            return ERR_NONE;
         int8 high = 1LL << (wsize - 1);
         int8 low = -high;
         high--;
@@ -1517,6 +1524,18 @@ int int2string(int4 n, char *buf, int buflen) {
     return count + uint2string(u, buf + count, buflen - count);
 }
 
+int ulong2string(uint8 n, char *buf, int buflen) {
+    uint8 pt = 1;
+    int count = 0;
+    while (n / pt >= 10)
+        pt *= 10;
+    while (pt != 0) {
+        char2buf(buf, buflen, &count, (char) ('0' + (n / pt) % 10));
+        pt /= 10;
+    }
+    return count;
+}
+
 int vartype2string(const vartype *v, char *buf, int buflen, int max_mant_digits) {
     int dispmode;
     int digits = 0;
