@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -17,6 +17,7 @@
 
 #include <stdlib.h>
 
+#include "core_globals.h"
 #include "core_linalg1.h"
 #include "core_linalg2.h"
 #include "core_main.h"
@@ -27,29 +28,29 @@
 /***** Matrix-matrix division *****/
 /**********************************/
 
-static void (*linalg_div_completion)(int, vartype *);
+static int (*linalg_div_completion)(int, vartype *);
 static const vartype *linalg_div_left;
 static vartype *linalg_div_result;
 
 static int div_rr_completion1(int error, vartype_realmatrix *a, int4 *perm,
                                     phloat det);
-static void div_rr_completion2(int error, vartype_realmatrix *a, int4 *perm,
+static int div_rr_completion2(int error, vartype_realmatrix *a, int4 *perm,
                                     vartype_realmatrix *b);
 static int div_rc_completion1(int error, vartype_complexmatrix *a, int4 *perm,
                                     phloat det_re, phloat det_im);
-static void div_rc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
+static int div_rc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
                                     vartype_complexmatrix *b);
 static int div_cr_completion1(int error, vartype_realmatrix *a, int4 *perm,
                                     phloat det);
-static void div_cr_completion2(int error, vartype_realmatrix *a, int4 *perm,
+static int div_cr_completion2(int error, vartype_realmatrix *a, int4 *perm,
                                     vartype_complexmatrix *b);
 static int div_cc_completion1(int error, vartype_complexmatrix *a, int4 *perm,
                                     phloat det_re, phloat det_im);
-static void div_cc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
+static int div_cc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
                                     vartype_complexmatrix *b);
 
 int linalg_div(const vartype *left, const vartype *right,
-                                    void (*completion)(int, vartype *)) {
+                                    int (*completion)(int, vartype *)) {
     if (left->type == TYPE_REALMATRIX) {
         if (right->type == TYPE_REALMATRIX) {
             vartype_realmatrix *num = (vartype_realmatrix *) left;
@@ -58,34 +59,28 @@ int linalg_div(const vartype *left, const vartype *right,
             int4 rows = num->rows;
             int4 columns = num->columns;
             int4 *perm;
-            if (denom->rows != rows || denom->columns != rows) {
-                completion(ERR_DIMENSION_ERROR, NULL);
-                return ERR_DIMENSION_ERROR;
-            }
+            if (denom->rows != rows || denom->columns != rows)
+                return completion(ERR_DIMENSION_ERROR, NULL);
             perm = (int4 *) malloc(rows * sizeof(int4));
-            if (perm == NULL) {
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
-            }
+            if (perm == NULL)
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             lu = new_realmatrix(rows, rows);
             if (lu == NULL) {
                 free(perm);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             res = new_realmatrix(rows, columns);
             if (res == NULL) {
                 free(perm);
                 free_vartype(lu);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             matrix_copy(lu, right);
             linalg_div_completion = completion;
             linalg_div_left = left;
             linalg_div_result = res;
             return lu_decomp_r((vartype_realmatrix *) lu, perm,
-                                                div_rr_completion1); 
+                                                div_rr_completion1);
         } else {
             vartype_realmatrix *num = (vartype_realmatrix *) left;
             vartype_complexmatrix *denom = (vartype_complexmatrix *) right;
@@ -93,27 +88,21 @@ int linalg_div(const vartype *left, const vartype *right,
             int4 rows = num->rows;
             int4 columns = num->columns;
             int4 *perm;
-            if (denom->rows != rows || denom->columns != rows) {
-                completion(ERR_DIMENSION_ERROR, NULL);
-                return ERR_DIMENSION_ERROR;
-            }
+            if (denom->rows != rows || denom->columns != rows)
+                return completion(ERR_DIMENSION_ERROR, NULL);
             perm = (int4 *) malloc(rows * sizeof(int4));
-            if (perm == NULL) {
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
-            }
+            if (perm == NULL)
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             lu = new_complexmatrix(rows, rows);
             if (lu == NULL) {
                 free(perm);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             res = new_complexmatrix(rows, columns);
             if (res == NULL) {
                 free(perm);
                 free_vartype(lu);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             matrix_copy(lu, right);
             linalg_div_completion = completion;
@@ -130,27 +119,21 @@ int linalg_div(const vartype *left, const vartype *right,
             int4 rows = num->rows;
             int4 columns = num->columns;
             int4 *perm;
-            if (denom->rows != rows || denom->columns != rows) {
-                completion(ERR_DIMENSION_ERROR, 0);
-                return ERR_DIMENSION_ERROR;
-            }
+            if (denom->rows != rows || denom->columns != rows)
+                return completion(ERR_DIMENSION_ERROR, 0);
             perm = (int4 *) malloc(rows * sizeof(int4));
-            if (perm == NULL) {
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
-            }
+            if (perm == NULL)
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             lu = new_realmatrix(rows, rows);
             if (lu == NULL) {
                 free(perm);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             res = new_complexmatrix(rows, columns);
             if (res == NULL) {
                 free(perm);
                 free_vartype(lu);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             matrix_copy(lu, right);
             linalg_div_completion = completion;
@@ -165,27 +148,21 @@ int linalg_div(const vartype *left, const vartype *right,
             int4 rows = num->rows;
             int4 columns = num->columns;
             int4 *perm;
-            if (denom->rows != rows || denom->columns != rows) {
-                completion(ERR_DIMENSION_ERROR, NULL);
-                return ERR_DIMENSION_ERROR;
-            }
+            if (denom->rows != rows || denom->columns != rows)
+                return completion(ERR_DIMENSION_ERROR, NULL);
             perm = (int4 *) malloc(rows * sizeof(int4));
-            if (perm == NULL) {
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
-            }
+            if (perm == NULL)
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             lu = new_complexmatrix(rows, rows);
             if (lu == NULL) {
                 free(perm);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             res = new_complexmatrix(rows, columns);
             if (res == NULL) {
                 free(perm);
                 free_vartype(lu);
-                completion(ERR_INSUFFICIENT_MEMORY, NULL);
-                return ERR_INSUFFICIENT_MEMORY;
+                return completion(ERR_INSUFFICIENT_MEMORY, NULL);
             }
             matrix_copy(lu, right);
             linalg_div_completion = completion;
@@ -212,13 +189,13 @@ static int div_rr_completion1(int error, vartype_realmatrix *a, int4 *perm,
     }
 }
 
-static void div_rr_completion2(int error, vartype_realmatrix *a, int4 *perm,
+static int div_rr_completion2(int error, vartype_realmatrix *a, int4 *perm,
                                           vartype_realmatrix *b) {
     if (error != ERR_NONE)
         free_vartype(linalg_div_result); /* Note: linalg_div_result == b */
     free_vartype((vartype *) a);
     free(perm);
-    linalg_div_completion(error, linalg_div_result);
+    return linalg_div_completion(error, linalg_div_result);
 }
 
 static int div_rc_completion1(int error, vartype_complexmatrix *a, int4 *perm,
@@ -236,13 +213,13 @@ static int div_rc_completion1(int error, vartype_complexmatrix *a, int4 *perm,
     }
 }
 
-static void div_rc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
+static int div_rc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
                                           vartype_complexmatrix *b) {
     if (error != ERR_NONE)
         free_vartype(linalg_div_result); /* Note: linalg_div_result == b */
     free_vartype((vartype *) a);
     free(perm);
-    linalg_div_completion(error, linalg_div_result);
+    return linalg_div_completion(error, linalg_div_result);
 }
 
 static int div_cr_completion1(int error, vartype_realmatrix *a, int4 *perm,
@@ -260,13 +237,13 @@ static int div_cr_completion1(int error, vartype_realmatrix *a, int4 *perm,
     }
 }
 
-static void div_cr_completion2(int error, vartype_realmatrix *a, int4 *perm,
+static int div_cr_completion2(int error, vartype_realmatrix *a, int4 *perm,
                                     vartype_complexmatrix *b) {
     if (error != ERR_NONE)
         free_vartype(linalg_div_result); /* Note: linalg_div_result == b */
     free_vartype((vartype *) a);
     free(perm);
-    linalg_div_completion(error, linalg_div_result);
+    return linalg_div_completion(error, linalg_div_result);
 }
 
 static int div_cc_completion1(int error, vartype_complexmatrix *a, int4 *perm,
@@ -284,13 +261,13 @@ static int div_cc_completion1(int error, vartype_complexmatrix *a, int4 *perm,
     }
 }
 
-static void div_cc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
+static int div_cc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
                                     vartype_complexmatrix *b) {
     if (error != ERR_NONE)
         free_vartype(linalg_div_result); /* Note: linalg_div_result == b */
     free_vartype((vartype *) a);
     free(perm);
-    linalg_div_completion(error, linalg_div_result);
+    return linalg_div_completion(error, linalg_div_result);
 }
 
 
@@ -298,21 +275,21 @@ static void div_cc_completion2(int error, vartype_complexmatrix *a, int4 *perm,
 /***** Matrix-matrix multiplication *****/
 /****************************************/
 
-typedef struct {
+struct mul_rr_data_struct {
     vartype_realmatrix *left;
     vartype_realmatrix *right;
     vartype *result;
     int4 i, j, k;
     phloat sum;
-    void (*completion)(int error, vartype *result);
-} mul_rr_data_struct;
+    int (*completion)(int error, vartype *result);
+};
 
 static mul_rr_data_struct *mul_rr_data;
 
-static int matrix_mul_rr_worker(int interrupted);
+static int matrix_mul_rr_worker(bool interrupted);
 
 static int matrix_mul_rr(vartype_realmatrix *left, vartype_realmatrix *right,
-                         void (*completion)(int, vartype *)) {
+                         int (*completion)(int, vartype *)) {
 
     mul_rr_data_struct *dat;
     int error;
@@ -322,7 +299,7 @@ static int matrix_mul_rr(vartype_realmatrix *left, vartype_realmatrix *right,
         goto finished;
     }
 
-    if (!contains_no_strings(left) || !contains_no_strings(right)) {
+    if (contains_strings(left) || contains_strings(right)) {
         error = ERR_ALPHA_DATA_IS_INVALID;
         goto finished;
     }
@@ -354,11 +331,10 @@ static int matrix_mul_rr(vartype_realmatrix *left, vartype_realmatrix *right,
     return ERR_INTERRUPTIBLE;
 
     finished:
-    completion(error, NULL);
-    return error;
+    return completion(error, NULL);
 }
 
-static int matrix_mul_rr_worker(int interrupted) {
+static int matrix_mul_rr_worker(bool interrupted) {
     mul_rr_data_struct *dat = mul_rr_data;
     int count = 0;
     int inf;
@@ -374,10 +350,10 @@ static int matrix_mul_rr_worker(int interrupted) {
     phloat sum = dat->sum;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, NULL);
+        int err = dat->completion(ERR_INTERRUPTED, NULL);
         free_vartype(dat->result);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     while (count++ < 1000) {
@@ -387,10 +363,10 @@ static int matrix_mul_rr_worker(int interrupted) {
         k = 0;
         if ((inf = p_isinf(sum)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
@@ -402,9 +378,9 @@ static int matrix_mul_rr_worker(int interrupted) {
         if (++i < m)
             continue;
         else {
-            dat->completion(ERR_NONE, dat->result);
+            int err = dat->completion(ERR_NONE, dat->result);
             free(dat);
-            return ERR_NONE;
+            return err;
         }
     }
 
@@ -453,7 +429,7 @@ static int matrix_mul_rr(vartype_realmatrix *left, vartype_realmatrix *right,
     m = left->rows;
     n = right->columns;
 
-    if (!contains_no_strings(left) || !contains_no_strings(right))
+    if (contains_strings(left) || contains_strings(right))
         return ERR_ALPHA_DATA_IS_INVALID;
 
     *result = new_realmatrix(m, n);
@@ -552,21 +528,21 @@ static int matrix_mul_rr(vartype_realmatrix *left, vartype_realmatrix *right,
 }
 #endif
 
-typedef struct {
+struct mul_rc_data_struct {
     vartype_realmatrix *left;
     vartype_complexmatrix *right;
     vartype *result;
     int4 i, j, k;
     phloat sum_re, sum_im;
-    void (*completion)(int error, vartype *result);
-} mul_rc_data_struct;
+    int (*completion)(int error, vartype *result);
+};
 
 static mul_rc_data_struct *mul_rc_data;
 
-static int matrix_mul_rc_worker(int interrupted);
+static int matrix_mul_rc_worker(bool interrupted);
 
 static int matrix_mul_rc(vartype_realmatrix *left, vartype_complexmatrix *right,
-                         void (*completion)(int, vartype *)) {
+                         int (*completion)(int, vartype *)) {
 
     mul_rc_data_struct *dat;
     int error;
@@ -576,7 +552,7 @@ static int matrix_mul_rc(vartype_realmatrix *left, vartype_complexmatrix *right,
         goto finished;
     }
 
-    if (!contains_no_strings(left)) {
+    if (contains_strings(left)) {
         error = ERR_ALPHA_DATA_IS_INVALID;
         goto finished;
     }
@@ -609,11 +585,10 @@ static int matrix_mul_rc(vartype_realmatrix *left, vartype_complexmatrix *right,
     return ERR_INTERRUPTIBLE;
 
     finished:
-    completion(error, NULL);
-    return error;
+    return completion(error, NULL);
 }
 
-static int matrix_mul_rc_worker(int interrupted) {
+static int matrix_mul_rc_worker(bool interrupted) {
     mul_rc_data_struct *dat = mul_rc_data;
     int count = 0;
     int inf;
@@ -630,10 +605,10 @@ static int matrix_mul_rc_worker(int interrupted) {
     phloat sum_im = dat->sum_im;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, NULL);
+        int err = dat->completion(ERR_INTERRUPTED, NULL);
         free_vartype(dat->result);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     while (count++ < 1000) {
@@ -645,19 +620,19 @@ static int matrix_mul_rc_worker(int interrupted) {
         k = 0;
         if ((inf = p_isinf(sum_re)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum_re = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
         if ((inf = p_isinf(sum_im)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum_im = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
@@ -671,9 +646,9 @@ static int matrix_mul_rc_worker(int interrupted) {
         if (++i < m)
             continue;
         else {
-            dat->completion(ERR_NONE, dat->result);
+            int err = dat->completion(ERR_NONE, dat->result);
             free(dat);
-            return ERR_NONE;
+            return err;
         }
     }
 
@@ -685,21 +660,21 @@ static int matrix_mul_rc_worker(int interrupted) {
     return ERR_INTERRUPTIBLE;
 }
 
-typedef struct {
+struct mul_cr_data_struct {
     vartype_complexmatrix *left;
     vartype_realmatrix *right;
     vartype *result;
     int4 i, j, k;
     phloat sum_re, sum_im;
-    void (*completion)(int error, vartype *result);
-} mul_cr_data_struct;
+    int (*completion)(int error, vartype *result);
+};
 
 static mul_cr_data_struct *mul_cr_data;
 
-static int matrix_mul_cr_worker(int interrupted);
+static int matrix_mul_cr_worker(bool interrupted);
 
 static int matrix_mul_cr(vartype_complexmatrix *left, vartype_realmatrix *right,
-                         void (*completion)(int, vartype *)) {
+                         int (*completion)(int, vartype *)) {
 
     mul_cr_data_struct *dat;
     int error;
@@ -709,7 +684,7 @@ static int matrix_mul_cr(vartype_complexmatrix *left, vartype_realmatrix *right,
         goto finished;
     }
 
-    if (!contains_no_strings(right)) {
+    if (contains_strings(right)) {
         error = ERR_ALPHA_DATA_IS_INVALID;
         goto finished;
     }
@@ -742,11 +717,10 @@ static int matrix_mul_cr(vartype_complexmatrix *left, vartype_realmatrix *right,
     return ERR_INTERRUPTIBLE;
 
     finished:
-    completion(error, NULL);
-    return error;
+    return completion(error, NULL);
 }
 
-static int matrix_mul_cr_worker(int interrupted) {
+static int matrix_mul_cr_worker(bool interrupted) {
     mul_cr_data_struct *dat = mul_cr_data;
     int count = 0;
     int inf;
@@ -763,10 +737,10 @@ static int matrix_mul_cr_worker(int interrupted) {
     phloat sum_im = dat->sum_im;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, NULL);
+        int err = dat->completion(ERR_INTERRUPTED, NULL);
         free_vartype(dat->result);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     while (count++ < 1000) {
@@ -778,19 +752,19 @@ static int matrix_mul_cr_worker(int interrupted) {
         k = 0;
         if ((inf = p_isinf(sum_re)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum_re = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
         if ((inf = p_isinf(sum_im)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum_im = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
@@ -804,9 +778,9 @@ static int matrix_mul_cr_worker(int interrupted) {
         if (++i < m)
             continue;
         else {
-            dat->completion(ERR_NONE, dat->result);
+            int err = dat->completion(ERR_NONE, dat->result);
             free(dat);
-            return ERR_NONE;
+            return err;
         }
     }
 
@@ -818,21 +792,21 @@ static int matrix_mul_cr_worker(int interrupted) {
     return ERR_INTERRUPTIBLE;
 }
 
-typedef struct {
+struct mul_cc_data_struct {
     vartype_complexmatrix *left;
     vartype_complexmatrix *right;
     vartype *result;
     int4 i, j, k;
     phloat sum_re, sum_im;
-    void (*completion)(int error, vartype *result);
-} mul_cc_data_struct;
+    int (*completion)(int error, vartype *result);
+};
 
 static mul_cc_data_struct *mul_cc_data;
 
-static int matrix_mul_cc_worker(int interrupted);
+static int matrix_mul_cc_worker(bool interrupted);
 
 static int matrix_mul_cc(vartype_complexmatrix *left, vartype_complexmatrix *right,
-                         void (*completion)(int, vartype *)) {
+                         int (*completion)(int, vartype *)) {
 
     mul_cc_data_struct *dat;
     int error;
@@ -870,11 +844,10 @@ static int matrix_mul_cc(vartype_complexmatrix *left, vartype_complexmatrix *rig
     return ERR_INTERRUPTIBLE;
 
     finished:
-    completion(error, NULL);
-    return error;
+    return completion(error, NULL);
 }
 
-static int matrix_mul_cc_worker(int interrupted) {
+static int matrix_mul_cc_worker(bool interrupted) {
     mul_cc_data_struct *dat = mul_cc_data;
     int count = 0;
     int inf;
@@ -891,10 +864,10 @@ static int matrix_mul_cc_worker(int interrupted) {
     phloat sum_im = dat->sum_im;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, NULL);
+        int err = dat->completion(ERR_INTERRUPTED, NULL);
         free_vartype(dat->result);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     while (count++ < 1000) {
@@ -909,19 +882,19 @@ static int matrix_mul_cc_worker(int interrupted) {
         k = 0;
         if ((inf = p_isinf(sum_re)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum_re = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
         if ((inf = p_isinf(sum_im)) != 0) {
             if (core_settings.matrix_outofrange && !flags.f.range_error_ignore){
-                dat->completion(ERR_OUT_OF_RANGE, NULL);
+                int err = dat->completion(ERR_OUT_OF_RANGE, NULL);
                 free_vartype(dat->result);
                 free(dat);
-                return ERR_OUT_OF_RANGE;
+                return err;
             } else
                 sum_im = inf < 0 ? NEG_HUGE_PHLOAT : POS_HUGE_PHLOAT;
         }
@@ -935,9 +908,9 @@ static int matrix_mul_cc_worker(int interrupted) {
         if (++i < m)
             continue;
         else {
-            dat->completion(ERR_NONE, dat->result);
+            int err = dat->completion(ERR_NONE, dat->result);
             free(dat);
-            return ERR_NONE;
+            return err;
         }
     }
 
@@ -950,7 +923,7 @@ static int matrix_mul_cc_worker(int interrupted) {
 }
 
 int linalg_mul(const vartype *left, const vartype *right,
-                                    void (*completion)(int, vartype *)) {
+                                    int (*completion)(int, vartype *)) {
     if (left->type == TYPE_REALMATRIX) {
         if (right->type == TYPE_REALMATRIX)
             return matrix_mul_rr((vartype_realmatrix *) left,
@@ -982,11 +955,11 @@ static vartype *linalg_inv_result;
 
 static int inv_r_completion1(int error, vartype_realmatrix *a, int4 *perm,
                                 phloat det);
-static void inv_r_completion2(int error, vartype_realmatrix *a, int4 *perm,
+static int inv_r_completion2(int error, vartype_realmatrix *a, int4 *perm,
                                 vartype_realmatrix *b);
 static int inv_c_completion1(int error, vartype_complexmatrix *a, int4 *perm,
                                 phloat det_re, phloat det_im);
-static void inv_c_completion2(int error, vartype_complexmatrix *a, int4 *perm,
+static int inv_c_completion2(int error, vartype_complexmatrix *a, int4 *perm,
                                 vartype_complexmatrix *b);
 
 int linalg_inv(const vartype *src, void (*completion)(int, vartype *)) {
@@ -998,7 +971,7 @@ int linalg_inv(const vartype *src, void (*completion)(int, vartype *)) {
         n = ma->rows;
         if (n != ma->columns)
             return ERR_DIMENSION_ERROR;
-        if (!contains_no_strings(ma))
+        if (contains_strings(ma))
             return ERR_ALPHA_DATA_IS_INVALID;
         lu = new_realmatrix(n, n);
         if (lu == NULL)
@@ -1063,13 +1036,14 @@ static int inv_r_completion1(int error, vartype_realmatrix *a, int4 *perm,
     }
 }
 
-static void inv_r_completion2(int error, vartype_realmatrix *a, int4 *perm,
+static int inv_r_completion2(int error, vartype_realmatrix *a, int4 *perm,
                                 vartype_realmatrix *b) {
     if (error != ERR_NONE)
         free_vartype(linalg_inv_result); /* Note: linalg_inv_result == b */
     free_vartype((vartype *) a);
     free(perm);
     linalg_inv_completion(error, linalg_inv_result);
+    return error;
 }
 
 static int inv_c_completion1(int error, vartype_complexmatrix *a, int4 *perm,
@@ -1090,13 +1064,14 @@ static int inv_c_completion1(int error, vartype_complexmatrix *a, int4 *perm,
     }
 }
 
-static void inv_c_completion2(int error, vartype_complexmatrix *a, int4 *perm,
+static int inv_c_completion2(int error, vartype_complexmatrix *a, int4 *perm,
                                 vartype_complexmatrix *b) {
     if (error != ERR_NONE)
         free_vartype(linalg_inv_result); /* Note: linalg_inv_result == b */
     free_vartype((vartype *) a);
     free(perm);
     linalg_inv_completion(error, linalg_inv_result);
+    return error;
 }
 
 
@@ -1122,7 +1097,7 @@ int linalg_det(const vartype *src, void (*completion)(int, vartype *)) {
             completion(ERR_DIMENSION_ERROR, 0);
             return ERR_DIMENSION_ERROR;
         }
-        if (!contains_no_strings(ma)) {
+        if (contains_strings(ma)) {
             completion(ERR_ALPHA_DATA_IS_INVALID, 0);
             return ERR_ALPHA_DATA_IS_INVALID;
         }
@@ -1154,7 +1129,7 @@ int linalg_det(const vartype *src, void (*completion)(int, vartype *)) {
         core_settings.matrix_singularmatrix = true;
 
         linalg_det_completion = completion;
-        return lu_decomp_r(ma, perm, det_r_completion); 
+        return lu_decomp_r(ma, perm, det_r_completion);
     } else /* src->type == TYPE_COMPLEXMATRIX */ {
         vartype_complexmatrix *ma = (vartype_complexmatrix *) src;
         n = ma->rows;
@@ -1185,7 +1160,7 @@ int linalg_det(const vartype *src, void (*completion)(int, vartype *)) {
         core_settings.matrix_singularmatrix = true;
 
         linalg_det_completion = completion;
-        return lu_decomp_c(ma, perm, det_c_completion); 
+        return lu_decomp_c(ma, perm, det_c_completion);
     }
 }
 
