@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -18,12 +18,12 @@
 #ifndef CORE_LINALG1_H
 #define CORE_LINALG1_H 1
 
-#include "core_globals.h"
+#include "core_variables.h"
 
 int linalg_div(const vartype *left, const vartype *right,
-                             void (*completion)(int, vartype *));
+                             int (*completion)(int, vartype *));
 int linalg_mul(const vartype *left, const vartype *right,
-                             void (*completion)(int, vartype *));
+                             int (*completion)(int, vartype *));
 int linalg_inv(const vartype *src, void (*completion)(int, vartype *));
 int linalg_det(const vartype *src, void (*completion)(int, vartype *));
 
