@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -35,7 +35,7 @@
 /***** LU decomposition *****/
 /****************************/
 
-typedef struct {
+struct lu_r_data_struct {
     vartype_realmatrix *a;
     int4 *perm;
     phloat det;
@@ -43,11 +43,11 @@ typedef struct {
     phloat max, tmp, sum, *scale;
     int state;
     int (*completion)(int, vartype_realmatrix *, int4 *, phloat);
-} lu_r_data_struct;
+};
 
 lu_r_data_struct *lu_r_data;
 
-static int lu_decomp_r_worker(int interrupted);
+static int lu_decomp_r_worker(bool interrupted);
 
 int lu_decomp_r(vartype_realmatrix *a, int4 *perm,
                 int (*completion)(int, vartype_realmatrix *, int4 *, phloat)) {
@@ -75,8 +75,8 @@ int lu_decomp_r(vartype_realmatrix *a, int4 *perm,
     return ERR_INTERRUPTIBLE;
 }
 
-static int lu_decomp_r_worker(int interrupted) {
-    
+static int lu_decomp_r_worker(bool interrupted) {
+
     lu_r_data_struct *dat = lu_r_data;
 
     phloat *a = dat->a->array->data;
@@ -218,7 +218,7 @@ static int lu_decomp_r_worker(int interrupted) {
 }
 
 
-typedef struct {
+struct lu_c_data_struct {
     vartype_complexmatrix *a;
     int4 *perm;
     phloat det_re, det_im;
@@ -226,11 +226,11 @@ typedef struct {
     phloat max, tmp, tmp_re, tmp_im, sum_re, sum_im, s_re, s_im, *scale;
     int state;
     int (*completion)(int, vartype_complexmatrix *, int4 *, phloat, phloat);
-} lu_c_data_struct;
+};
 
 lu_c_data_struct *lu_c_data;
 
-static int lu_decomp_c_worker(int interrupted);
+static int lu_decomp_c_worker(bool interrupted);
 
 int lu_decomp_c(vartype_complexmatrix *a, int4 *perm,
                 int (*completion)(int, vartype_complexmatrix *,
@@ -259,8 +259,8 @@ int lu_decomp_c(vartype_complexmatrix *a, int4 *perm,
     return ERR_INTERRUPTIBLE;
 }
 
-static int lu_decomp_c_worker(int interrupted) {
-    
+static int lu_decomp_c_worker(bool interrupted) {
+
     lu_c_data_struct *dat = lu_c_data;
 
     phloat *a = dat->a->array->data;
@@ -335,6 +335,7 @@ static int lu_decomp_c_worker(int interrupted) {
         }
 
         max = 0;
+        imax = j;
         for (i = j; i < n; i++) {
             sum_re = a[2 * (i * n + j)];
             sum_im = a[2 * (i * n + j) + 1];
@@ -444,30 +445,28 @@ static int lu_decomp_c_worker(int interrupted) {
 /***** Back-substitution *****/
 /*****************************/
 
-typedef struct {
+struct backsub_rr_data_struct {
     vartype_realmatrix *a;
     int4 *perm;
     vartype_realmatrix *b;
     int4 i, ii, j, ll, k;
     phloat sum;
     int state;
-    void (*completion)(int, vartype_realmatrix *, int4 *, vartype_realmatrix *);
-} backsub_rr_data_struct;
+    int (*completion)(int, vartype_realmatrix *, int4 *, vartype_realmatrix *);
+};
 
 static backsub_rr_data_struct *backsub_rr_data;
 
-static int lu_backsubst_rr_worker(int interrupted);
+static int lu_backsubst_rr_worker(bool interrupted);
 
 int lu_backsubst_rr(vartype_realmatrix *a, int4 *perm, vartype_realmatrix *b,
-                    void (*completion)(int, vartype_realmatrix *,
+                    int (*completion)(int, vartype_realmatrix *,
                                     int4 *, vartype_realmatrix *)) {
     backsub_rr_data_struct *dat =
             (backsub_rr_data_struct *) malloc(sizeof(backsub_rr_data_struct));
 
-    if (dat == NULL) {
-        completion(ERR_INSUFFICIENT_MEMORY, a, perm, b);
-        return ERR_INSUFFICIENT_MEMORY;
-    }
+    if (dat == NULL)
+        return completion(ERR_INSUFFICIENT_MEMORY, a, perm, b);
 
     dat->a = a;
     dat->perm = perm;
@@ -482,7 +481,7 @@ int lu_backsubst_rr(vartype_realmatrix *a, int4 *perm, vartype_realmatrix *b,
     return ERR_INTERRUPTIBLE;
 }
 
-static int lu_backsubst_rr_worker(int interrupted) {
+static int lu_backsubst_rr_worker(bool interrupted) {
     backsub_rr_data_struct *dat = backsub_rr_data;
     phloat *a = dat->a->array->data;
     int4 n = dat->a->rows;
@@ -501,9 +500,9 @@ static int lu_backsubst_rr_worker(int interrupted) {
     phloat t;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, dat->a, perm, dat->b);
+        int err = dat->completion(ERR_INTERRUPTED, dat->a, perm, dat->b);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     switch (dat->state) {
@@ -545,9 +544,10 @@ static int lu_backsubst_rr_worker(int interrupted) {
         }
     }
 
-    dat->completion(ERR_NONE, dat->a, perm, dat->b);
+    int err;
+    err = dat->completion(ERR_NONE, dat->a, perm, dat->b);
     free(dat);
-    return ERR_NONE;
+    return err;
 
     suspend:
     dat->i = i;
@@ -559,31 +559,29 @@ static int lu_backsubst_rr_worker(int interrupted) {
     return ERR_INTERRUPTIBLE;
 }
 
-typedef struct {
+struct backsub_rc_data_struct {
     vartype_realmatrix *a;
     int4 *perm;
     vartype_complexmatrix *b;
     int4 i, ii, j, ll, k;
     phloat sum_re, sum_im;
     int state;
-    void (*completion)(int, vartype_realmatrix *, int4 *,
+    int (*completion)(int, vartype_realmatrix *, int4 *,
                                             vartype_complexmatrix *);
-} backsub_rc_data_struct;
+};
 
 static backsub_rc_data_struct *backsub_rc_data;
 
-static int lu_backsubst_rc_worker(int interrupted);
+static int lu_backsubst_rc_worker(bool interrupted);
 
 int lu_backsubst_rc(vartype_realmatrix *a, int4 *perm, vartype_complexmatrix *b,
-                    void (*completion)(int, vartype_realmatrix *,
+                    int (*completion)(int, vartype_realmatrix *,
                                 int4 *, vartype_complexmatrix *)) {
     backsub_rc_data_struct *dat =
             (backsub_rc_data_struct *) malloc(sizeof(backsub_rc_data_struct));
 
-    if (dat == NULL) {
-        completion(ERR_INSUFFICIENT_MEMORY, a, perm, b);
-        return ERR_INSUFFICIENT_MEMORY;
-    }
+    if (dat == NULL)
+        return completion(ERR_INSUFFICIENT_MEMORY, a, perm, b);
 
     dat->a = a;
     dat->perm = perm;
@@ -598,7 +596,7 @@ int lu_backsubst_rc(vartype_realmatrix *a, int4 *perm, vartype_complexmatrix *b,
     return ERR_INTERRUPTIBLE;
 }
 
-static int lu_backsubst_rc_worker(int interrupted) {
+static int lu_backsubst_rc_worker(bool interrupted) {
     backsub_rc_data_struct *dat = backsub_rc_data;
     phloat *a = dat->a->array->data;
     int4 n = dat->a->rows;
@@ -619,9 +617,9 @@ static int lu_backsubst_rc_worker(int interrupted) {
     phloat t_re, t_im;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, dat->a, perm, dat->b);
+        int err = dat->completion(ERR_INTERRUPTED, dat->a, perm, dat->b);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     switch (dat->state) {
@@ -681,9 +679,10 @@ static int lu_backsubst_rc_worker(int interrupted) {
         }
     }
 
-    dat->completion(ERR_NONE, dat->a, perm, dat->b);
+    int err;
+    err = dat->completion(ERR_NONE, dat->a, perm, dat->b);
     free(dat);
-    return ERR_NONE;
+    return err;
 
     suspend:
     dat->i = i;
@@ -696,31 +695,29 @@ static int lu_backsubst_rc_worker(int interrupted) {
     return ERR_INTERRUPTIBLE;
 }
 
-typedef struct {
+struct backsub_cc_data_struct {
     vartype_complexmatrix *a;
     int4 *perm;
     vartype_complexmatrix *b;
     int4 i, ii, j, ll, k;
     phloat sum_re, sum_im;
     int state;
-    void (*completion)(int, vartype_complexmatrix *, int4 *,
+    int (*completion)(int, vartype_complexmatrix *, int4 *,
                                             vartype_complexmatrix *);
-} backsub_cc_data_struct;
+};
 
 static backsub_cc_data_struct *backsub_cc_data;
 
-static int lu_backsubst_cc_worker(int interrupted);
+static int lu_backsubst_cc_worker(bool interrupted);
 
 int lu_backsubst_cc(vartype_complexmatrix *a, int4 *perm, vartype_complexmatrix *b,
-                    void (*completion)(int, vartype_complexmatrix *,
+                    int (*completion)(int, vartype_complexmatrix *,
                                 int4 *, vartype_complexmatrix *)) {
     backsub_cc_data_struct *dat =
             (backsub_cc_data_struct *) malloc(sizeof(backsub_cc_data_struct));
 
-    if (dat == NULL) {
-        completion(ERR_INSUFFICIENT_MEMORY, a, perm, b);
-        return ERR_INSUFFICIENT_MEMORY;
-    }
+    if (dat == NULL)
+        return completion(ERR_INSUFFICIENT_MEMORY, a, perm, b);
 
     dat->a = a;
     dat->perm = perm;
@@ -735,7 +732,7 @@ int lu_backsubst_cc(vartype_complexmatrix *a, int4 *perm, vartype_complexmatrix
     return ERR_INTERRUPTIBLE;
 }
 
-static int lu_backsubst_cc_worker(int interrupted) {
+static int lu_backsubst_cc_worker(bool interrupted) {
     backsub_cc_data_struct *dat = backsub_cc_data;
     phloat *a = dat->a->array->data;
     int4 n = dat->a->rows;
@@ -757,9 +754,9 @@ static int lu_backsubst_cc_worker(int interrupted) {
     phloat t_re, t_im;
 
     if (interrupted) {
-        dat->completion(ERR_INTERRUPTED, dat->a, perm, dat->b);
+        int err = dat->completion(ERR_INTERRUPTED, dat->a, perm, dat->b);
         free(dat);
-        return ERR_INTERRUPTED;
+        return err;
     }
 
     switch (dat->state) {
@@ -829,9 +826,10 @@ static int lu_backsubst_cc_worker(int interrupted) {
         }
     }
 
-    dat->completion(ERR_NONE, dat->a, perm, dat->b);
+    int err;
+    err = dat->completion(ERR_NONE, dat->a, perm, dat->b);
     free(dat);
-    return ERR_NONE;
+    return err;
 
     suspend:
     dat->i = i;
