@@ -1,6 +1,6 @@
 /*****************************************************************************
  * Free42 -- an HP-42S calculator simulator
- * Copyright (C) 2004-2021  Thomas Okken
+ * Copyright (C) 2004-2022  Thomas Okken
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License, version 2,
@@ -72,7 +72,7 @@ void core_repaint_display();
  * whether or not clicking in the display, to activate menu keys, is
  * enabled.
  */
-int core_menu();
+bool core_menu();
 
 /* core_alpha_menu()
  *
@@ -83,7 +83,7 @@ int core_menu();
  * straight to the core; outside alpha mode, all key events are translated
  * sequences of HP-42S key events according to the keymap file.
  */
-int core_alpha_menu();
+bool core_alpha_menu();
 
 /* core_hex_menu()
  *
@@ -93,7 +93,18 @@ int core_alpha_menu();
  * Free42) are handled: in hex mode, 'A' through 'F' and 'a' through 'f' are
  * translated to keycodes 1 through 6, regardless of the keyboard map.
  */
-int core_hex_menu();
+bool core_hex_menu();
+
+/* core_special_menu_key()
+ *
+ * The shell can call this function if the physical cursor left, cursor right,
+ * or delete keys have been pressed; the core will return the appropriate
+ * menu key to trigger. This is to support menus like the matrix editor.
+ * Which can be: 1=left, 2=shift left, 3=right, 4=shift right, 5=del.
+ * The return value is a menu key number 1-6, or 0 if no special menu key
+ * is available.
+ */
+int core_special_menu_key(int which);
 
 /* core_keydown()
  *
@@ -129,7 +140,7 @@ int core_hex_menu();
  * core_repeat() until the key is released. (1 requests a slow repeat rate, for
  * SST/BST; 2 requests a fast repeat rate, for number/alpha entry.)
  */
-int core_keydown(int key, int *enqueued, int *repeat);
+bool core_keydown(int key, bool *enqueued, int *repeat);
 
 /* core_keydown_command()
  *
@@ -142,7 +153,7 @@ int core_keydown(int key, int *enqueued, int *repeat);
  * name, or selects the command from a menu or the FCN catalog, all of which
  * have potentially undesirable side effects.
  */
-int core_keydown_command(const char *name, int *enqueued, int *repeat);
+bool core_keydown_command(const char *name, bool *enqueued, int *repeat);
 
 /* core_repeat()
  *
@@ -186,7 +197,7 @@ void core_keytimeout2();
  * If the function returns 'true', this means that the timeout was used for
  * PSE, and the shell should resume program execution.
  */
-bool core_timeout3(int repaint);
+bool core_timeout3(bool repaint);
 
 /* core_keyup()
  *
@@ -202,7 +213,7 @@ bool core_timeout3(int repaint);
  * running a user program, and is only returning execution to the shell because
  * it has detected that there is a pending event.
  */
-int core_keyup();
+bool core_keyup();
 
 /* core_powercycle()
  *
@@ -219,7 +230,7 @@ int core_keyup();
  * to run again -- this will be true if flag 11 was set, so there's program
  * execution to be done.
  */
-int core_powercycle();
+bool core_powercycle();
 
 /* core_list_programs()
  *
@@ -256,7 +267,7 @@ int4 core_program_size(int prgm_index);
  *
  * This function is called by the shell after the user has selected a nonempty
  * set of programs (from the list returned by core_list_programs()) and
- * confirmed the operation (supplied a file name etc.). 
+ * confirmed the operation (supplied a file name etc.).
  * The 'count' parameter indicates how many programs are to be exported; the
  * 'indexes' parameter is an array of program indexes.
  * When called by the core during state file saving, raw_file_name will be
@@ -276,7 +287,9 @@ void core_import_programs(int num_progs, const char *raw_file_name);
 
 /* core_copy()
  *
- * Returns a string representation of the contents of the X register.
+ * Returns a string representation of the contents of the X register. In
+ * program mode, returns a listing of the current program. I alpha mode,
+ * returns the contents of the alpha register.
  * Used by the shell to implement the Copy command.
  * The caller should free the returned text using free(3).
  */
@@ -284,24 +297,36 @@ char *core_copy();
 
 /* core_paste()
  *
- * Puts the given value on the stack, using RCL semantics. It tries to parse
- * the string as a complex or a real number; if that fails, it is pasted as
- * a plain string.
+ * In normal mode, puts the given value on the stack, parsing it as tab-
+ * delimited spreadsheet cells, complex or real scalars, or, if nothing
+ * matches, as a simple string. In program mode, attempts to parse the text as
+ * a program listing, and adds the converted code as a new program at the end
+ * of program memory. In alpha mode, pastes the text in the alpha register,
+ * appending it to whatever is already there.
  * Used by the shell to implement the Paste command.
  */
 void core_paste(const char *s);
 
+/* core_update_allow_big_stack()
+ *
+ * Updates the big stack state and the UI to reflect a change in the
+ * global Allow Big Stack setting.
+ */
+void core_update_allow_big_stack();
+
 /* core_settings
  *
  * This is a struct that stores user-configurable core settings. The shell
  * should provide the appropriate controls in a "Preferences" dialog box to
  * allow the user to view and change these settings.
  */
-typedef struct {
+struct core_settings_struct {
     bool matrix_singularmatrix;
     bool matrix_outofrange;
     bool auto_repeat;
-} core_settings_struct;
+    bool allow_big_stack;
+    bool localized_copy_paste;
+};
 
 extern core_settings_struct core_settings;
 
@@ -326,7 +351,7 @@ bool alpha_active();
 
 int want_to_run_again();
 void do_interactive(int command);
-int find_builtin(const char *name, int namelen, bool strict);
+int find_builtin(const char *name, int namelen);
 
 void sst();
 void bst();
@@ -340,6 +365,7 @@ void start_alpha_prgm_line();
 void finish_alpha_prgm_line();
 int shiftcharacter(char c);
 void set_old_pc(int4 pc);
+const char *number_format();
 
 
 #endif
