@@ -176,7 +176,8 @@ extern phloat NAN_2_PHLOAT;
 void phloat_init();
 int phloat2string(phloat d, char *buf, int buflen,
                   int base_mode, int digits, int dispmode,
-                  int thousandssep, int max_mant_digits = 12);
+                  int thousandssep, int max_mant_digits = 12,
+                  const char *format = NULL);
 int string2phloat(const char *buf, int buflen, phloat *d);
 
 
