@@ -40,7 +40,7 @@ void skin_put_pixels(unsigned const char *data);
 void skin_finish_image();
 
 void skin_repaint(cairo_t *cr);
-void skin_repaint_annunciator(cairo_t *cr, int which, bool state);
+void skin_repaint_annunciator(cairo_t *cr, int which);
 void skin_invalidate_annunciator(GdkWindow *win, int which);
 void skin_find_key(int x, int y, bool cshift, int *key, int *code);
 int skin_find_skey(int ckey);
@@ -52,6 +52,7 @@ void skin_repaint_key(cairo_t *cr, int key, bool state);
 void skin_invalidate_key(GdkWindow *win, int key);
 void skin_display_invalidater(GdkWindow *win, const char *bits, int bytesperline,
                                 int x, int y, int width, int height);
+bool need_to_paint_only_display(cairo_t *cr);
 void skin_repaint_display(cairo_t *cr);
 void skin_invalidate_display(GdkWindow *win);
 void skin_display_set_enabled(bool enable);
